/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 *    "Apache Lucene" must not be used to endorse or promote products
 *    derived from this software without prior written permission. For
 *    written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    "Apache Lucene", nor may "Apache" appear in their name, without
 *    prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
*/

package org.apache.lucene.analysis.cjk;

import org.apache.lucene.analysis.Tokenizer;
import org.apache.lucene.analysis.Token;
import java.io.Reader;

/**
 * CJKTokenizer was modified from StopTokenizer which does a decent job for most European
 * languages. and it perferm other token method for double-byte Characters: the token will
 * return at each two charactors with overlap match.
 * Example: "java C1C2C3C4" will be segment to: "java" "C1C2" "C2C3" "C3C4"
 * it also need filter filter zero length token ""
 *
 * for more info on Asia language(Chinese Japanese Korean) text segmentation:
 * http://www.google.com/search?q=overlap+match+chinese+segment
 * for Digit: the prefix digit will token: "3dmax"=>"3" "dmax"; "U2"=>"u2"
 * for Punc: '_' will token as a letter, '+' '#' will token as a digit
 *
 * @author    Che, Dong <chedong@bigfoot.com>
 * $Header: /home/cvsroot/lucene_ext/src/org/apache/lucene/analysis/cjk/CJKTokenizer.java,v 1.1.1.1 2002/09/22 19:36:08 chedong Exp $
 */

public final class CJKTokenizer extends Tokenizer {
    private int offset = 0, bufferIndex=0, dataLen=0;
    private final static int MAX_WORD_LEN = 255;
    private final static int IO_BUFFER_SIZE = 256;//
    private final char[] buffer = new char[MAX_WORD_LEN];
    private final char[] ioBuffer = new char[IO_BUFFER_SIZE];
    private String tokenType = "word"; //single=>ASCII  double=>non-ASCII word=>default
    private boolean preIsTokened = false;//previous character is a cached double-byte character

    public CJKTokenizer(Reader in) {
        input = in;
    }

    public final Token next() throws java.io.IOException {
        int length = 0;
        int start = offset;
        while (true) {
            char c;
            Character.UnicodeBlock ub;

            offset++;
            if (bufferIndex >= dataLen) {
                dataLen = input.read(ioBuffer);
                bufferIndex = 0;
            }
            if (dataLen == -1) {
                if (length > 0){
                    if (preIsTokened == true){
                        length = 0;//empty buffer
                        preIsTokened = false;
                    }
                    break;
                }
                else
                    return null;
            }
            else {
                c = (char) ioBuffer[bufferIndex++];
                ub = Character.UnicodeBlock.of(c);
            }

            //ASCII and Extend ASCII
            if (ub == Character.UnicodeBlock.BASIC_LATIN
                    || ub == Character.UnicodeBlock.HALFWIDTH_AND_FULLWIDTH_FORMS ) {

                //convert  HALFWIDTH_AND_FULLWIDTH_FORMS to BASIC_LATIN
                if  (ub == Character.UnicodeBlock.HALFWIDTH_AND_FULLWIDTH_FORMS) {
                    int i  = (int) c ;
                    i = i - 65248;
                    c = (char) i;
                }

                // if it's a letter or "_"
                if (Character.isLetter(c) || c == '_') {
                    if (length == 0) { // start of token
                        start = offset - 1;
                    }
                    else if (tokenType == "double") {
                        offset --;
                        bufferIndex --;
                        tokenType = "single";
                        if (preIsTokened == true) {
                            length = 0;//empty buffer
                            preIsTokened = false;
                            break;
                        }
                        else {
                            break;
                        }
                    }
                    else if (tokenType == "digit"){
                        offset --;
                        bufferIndex --;
                        break;
                    }

                    buffer[length++] = Character.toLowerCase(c);  // buffer it
                    tokenType = "single";
                    if (length == MAX_WORD_LEN) // buffer overflow!
                        break;
                }
                else if  (Character.isDigit(c) || c == '+'  ||  c == '#') {
                    if (length == 0) { // start of token
                        if (c == '+'  ||  c == '#') {
                            //ignore '++i'
                            break;
                        }
                        start = offset - 1;
                    }
                    else if (tokenType == "double") {
                        offset --;
                        bufferIndex --;
                        tokenType = "digit";
                        if (preIsTokened == true){
                            length = 0;//empty buffer
                            preIsTokened = false;
                            break;
                        }
                        else {
                            break;
                        }
                    }

                    tokenType = "digit";
                    buffer[length++] = c;// buffer it

                    if (length == MAX_WORD_LEN)// buffer overflow!
                        break;
                }
                else if (length > 0) {			  // at non-Letter w/ chars
                    if (preIsTokened == true) {
                        length = 0;//empty buffer
                        preIsTokened = false;
                    }
                    else {
                        break;
                    }
                }
            }
            //non-ASCII
            else {
                if (Character.isLetter(c)) {// if it's a letter
                    if (length == 0){
                        start = offset -1 ;
                        buffer[length++] = c;
                        tokenType = "double";
                    }
                    else{
                        if (tokenType == "single" || tokenType == "digit"){
                            offset --;
                            bufferIndex--;
                            break;
                        }
                        else{
                            buffer[length++] = c;
                            tokenType = "double";
                            if (length == 2){
                                offset --;
                                bufferIndex--;
                                preIsTokened = true;
                                break;
                            }
                        }
                    }
                }
                else if (length > 0){
                    if (preIsTokened == true){
                        length = 0;//empty buffer
                        preIsTokened = false;
                    }
                    else{
                        break;
                    }
                }
            }
        }
        return new Token(new String(buffer, 0, length), start, start+length, tokenType);
    }
}

