package org.apache.lucene.demo;

/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 *    "Apache Lucene" must not be used to endorse or promote products
 *    derived from this software without prior written permission. For
 *    written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    "Apache Lucene", nor may "Apache" appear in their name, without
 *    prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.

 */

import java.util.StringTokenizer;
import java.util.Enumeration;
import java.io.File;
import java.io.IOException;

import org.xml.sax.*;

import org.apache.lucene.document.Field;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexWriter;

/**
 * Use Sax reader read xml source and built lucene index.
 * the xml source format as following dtd:
<!--
- Top level: table
-->
<!ELEMENT Records (Record*)>
    <!--
     - second level: record
    -->
    <!ELEMENT Record (Field*, Index*)>

        <!--
         - third level: fields
        -->
        <!ELEMENT Field (#PCDATA)>
            <!--field name-->
            <!ATTLIST Field name CDATA #REQUIRED>
            <!--field store tag: "yes"/"no" default "yes" -->
            <!ATTLIST Field store (yes|no) "yes">

        <!--
         - third level: index
        -->
        <!ELEMENT Index (#PCDATA)>
            <!--field name-->
            <!ATTLIST Index name CDATA #REQUIRED>
            <!--field token tag: "yes"/"no" default "yes" -->
            <!ATTLIST Index token (yes|no) "yes">

 *
 * $Header: /home/cvsroot/lucene_ext/src/org/apache/lucene/demo/XMLIndexer.java,v 1.1.1.1 2002/09/22 19:36:08 chedong Exp $
 */
public final class XMLIndexer implements ContentHandler, ErrorHandler {
    private XMLReader xr = null; //xml sax reader
    private IndexWriter iw = null;  //IndexWriter

    //trace info
    private int docTotal = 0;
    public int LOG_REPORT_CONT = 1000; //default log report interval
    private int reportCounter = LOG_REPORT_CONT;
    long startTime = System.currentTimeMillis(); //timer
    long endTime = 0;

    private String fieldName = null;  //field name
    private String fieldValue = null; //field value
    private boolean storeTag = false; //need store
    private boolean indexTag = false; //need index
    private boolean tokenTag = false; //need token

    private Field currentField = null;
    private Document currentDoc = new Document();

    private byte currentLevel = 0; //locate XML tree level depth: 1 2 3

    /**
     * create XMLIndexer: xml sax reader and lucene index writer
     */
    public XMLIndexer(XMLReader xmlReader,  IndexWriter indexWriter) {
        xr = xmlReader;
        iw = indexWriter;
        //set content handler
        xr.setContentHandler(this);
        //set error handler
        xr.setErrorHandler(this);
    }

    /**
     * Returns whether lucene index build successful complete
     * @param  src: the xml input source.
     * @return boolean: if build successful complete return true else return false.
     */
    public boolean buildIndex(InputSource src) throws  SAXException {
        try {
            if ( (iw != null) && (src != null) ) {
                xr.parse( src );
                iw.optimize();
                iw.close();
                endTime = System.currentTimeMillis();
                System.out.println(docTotal + " rows added\tTotal time Use:"
                                   + (endTime - startTime)/1000 + " second");
            }
            else {
                return false;
            }
            return true;
        }
        catch (SAXException se) {
            System.out.println( "failed with SAX error: " + se.toString() );
            se.printStackTrace();
            return false;
        }
        catch (IOException ioe) {
            System.out.println( "failed with IO error: " + ioe.getMessage() );
            ioe.printStackTrace();
            return false;
        }
    }

    /**
     * Returns whether lucene index build successful complete
     * @param  src: the xml source.
     * @param  dest: index writer
     * @param  c: index counter, each number of docs add write a log
     * @return boolean: if build successful complete return true else return false.
     */
    public boolean buildIndex(InputSource src, int c) throws  SAXException {
        LOG_REPORT_CONT = c;
        return buildIndex(src);
    }

    ////////////////////////////////////////////////////////////////////
    // Implementation of org.xml.sax.ContentHandler.
    ////////////////////////////////////////////////////////////////////

    public void setDocumentLocator (Locator locator) {	}

    public void startDocument () throws SAXException {
        //init Counter
        docTotal = 0;
        //start at root level
        currentLevel = 0;
        //reset report Counter
        reportCounter = LOG_REPORT_CONT;
    }

    public void endDocument () throws SAXException {	}

    public void startPrefixMapping (String prefix, String uri) throws SAXException {	}

    public void endPrefixMapping (String prefix) throws SAXException {	}

    public void startElement (String namespaceURI, String localName,
                              String qName, Attributes atts) throws SAXException {
        //to sub level
        currentLevel ++;
        switch ( currentLevel ){
        case 1:    //table level
            break;
        case 2:    //record level
            currentDoc = new Document();
            break;
        case 3:    //field level
            try {
                if (atts.getValue("name") != null ) {
                    fieldName = new String(atts.getValue("name")).trim();
                    //default values
                    fieldValue = new String();
                    storeTag = false;
                    indexTag = false;
                    tokenTag = false;

                    if ( localName.equals("Field") ) {
                        String store = atts.getValue("store") ;
                        if ( store != null && store.equals("no") ) {
                            storeTag = false;
                        }
                        else {
                            //default store
                            storeTag = true;
                        }
                    }
                    else if ( localName.equals("Index") ) {
                        indexTag = true;
                        String token = atts.getValue("token");
                        if  ( token != null && token.equals("no") ) {
                            tokenTag = false;
                        }
                        else {
                            tokenTag = true;
                        }
                    }
                }
            }
            catch (Exception e) {
                System.out.println(e.toString());
            }
            break;
        }
    }

    public void endElement (String namespaceURI, String localName,String qName) throws SAXException {
        switch ( currentLevel ) {
        case 1:  //table level
            break;
        case 2:  //record level
            //mapping field to index:
            try {
                currentDoc = mapDoc(currentDoc);
                //write to document
                iw.addDocument( currentDoc );
                //counter ++
                docTotal ++;
                reportCounter --;
                if ( reportCounter == 0) {
                    //show status;
                    endTime = System.currentTimeMillis();
                    System.out.println(docTotal + " rows added\ttime Use:"
                                       + (endTime - startTime)/1000 + " second");
                    //reset reportCounter
                    reportCounter = LOG_REPORT_CONT;
                }
            }
            catch (Exception e) {
                System.out.println(e.toString());
            }
            break;
        case 3:  //field level
            if ( fieldName != null && fieldName.length() > 0 ) {
                currentField = new Field(fieldName, fieldValue,
                                         storeTag, indexTag, tokenTag);
                currentDoc.add(currentField);
            }
            break;
        }
        //back to up level
        currentLevel --;
    }

    public void characters (char ch[], int start, int length) throws SAXException {
        //read field value
        if ( currentLevel == 3 ) {
            /* NOTICE:
             * if use: fieldValue = new String(ch, start, length)
             * may cause xml data value broken during saxReader reaches buffer end
             * for example:
             *                      <SomeTag>my content</SomeTag>
             * privous sax buffer reached here---^
             * after next buffer read will invoke another characters() event
             * so the fieldValue will return broken value "ntent" only
             */
            fieldValue = fieldValue + new String(ch, start, length);
        }
    }

    public void ignorableWhitespace (char ch[], int start, int length) throws SAXException { }

    public void processingInstruction (String target, String data) throws SAXException { }

    public void skippedEntity (String name) throws SAXException { }

    ////////////////////////////////////////////////////////////////////
    // Implementation of org.xml.sax.ErrorHandler.
    ////////////////////////////////////////////////////////////////////

    public void warning (SAXParseException e) throws SAXException  {
        System.out.println("  EVENT: warning " +   e.getMessage() + ' ' +
                           e.getSystemId() + ' ' +  e.getLineNumber() + ' ' +
                           e.getColumnNumber());
    }

    public void error (SAXParseException e) throws SAXException  {
        System.out.println("  EVENT: error " +   e.getMessage() + ' ' +
                           e.getSystemId() + ' ' +  e.getLineNumber() + ' ' +
                           e.getColumnNumber());
    }

    public void fatalError (SAXParseException e) throws SAXException {
        System.out.println("  EVENT: fatal error " +
                           e.getMessage() + ' ' +   e.getSystemId() + ' ' +
                           e.getLineNumber() + ' ' +  e.getColumnNumber());
    }

    /**
     * map parsed field document to lucene index mapped field
     * @return boolean: if build successful complete return true else return false.
     */
    private static Document mapDoc(Document parsedDoc) {
        Document newDoc = new Document();
        try {
            Enumeration fe = parsedDoc.fields();
            while ( fe.hasMoreElements() ){
                Field fld = (Field)fe.nextElement();

                //index map field with fields name list: 'field1,field2,field5....'
                if ( fld.isIndexed() ) {
                    String indexName = fld.name();
                    StringBuffer indexValue = new StringBuffer();

                    String fieldList =  fld.stringValue();
                    StringTokenizer st = new StringTokenizer(fieldList,",");
                    while ( st.hasMoreTokens() ) {
                        //add indexValue with mapped field value
                        String mapFieldName = new String();
                        mapFieldName = st.nextToken();
                        Field mapField = parsedDoc.getField(mapFieldName);
                        String mapValue = null;

                        if ( mapField != null ) {
                            mapValue = mapField.stringValue();
                        }

                        if ( mapValue != null ) {
                            indexValue.append( mapValue );
                            //add \t between field values avoid "field1field2"
                            indexValue.append( "\t" );
                        }
                    }

                    if (indexValue.length() > 0) {
                        Field newIndex = new Field(indexName, indexValue.toString(),
                                                   false, true, fld.isTokenized() );
                        newDoc.add(newIndex);
                    }
                }
                //add common field
                else {
                    newDoc.add(fld);
                }
            }
        }
        catch (Exception e) {
            System.out.println(e.toString());
            return null;
        }
        return newDoc;
    }
}
